import ast
from copy import deepcopy
from dataclasses import fields, is_dataclass
from enum import Enum, auto
import inspect
from typing import Any, Callable, Iterable, Iterator, Optional, TypeAliasType


class AnnotationLevel(Enum):
    NONE = auto()
    WRONG = auto()
    CORRECT = auto()

    def __bool__(self) -> bool:
        match self:
            case AnnotationLevel.NONE | AnnotationLevel.WRONG:
                return False
            case AnnotationLevel.CORRECT:
                return True

    __nonzero__ = __bool__


def has_annotation_function(f: Callable, param_types: list[type], return_type: object = None) -> AnnotationLevel:
    a = f.__annotations__
    if len(a) == 0:
        return AnnotationLevel.NONE

    ret = a['return'] if 'return' in a else None
    if ret != return_type:
        return AnnotationLevel.WRONG

    params = list(a.values())[:-1] if 'return' in a else list(a.values())
    for param, param_type in zip(params, param_types):
        while isinstance(param, TypeAliasType):
            param = param.__value__
        while isinstance(param_type, TypeAliasType):
            param_type = param_type.__value__
        if param != param_type:
            return AnnotationLevel.WRONG
    return AnnotationLevel.CORRECT


def has_annotation_dataclass(cls: object, types: list[object]) -> AnnotationLevel:
    if not is_dataclass(cls):
        AnnotationLevel.WRONG
    types_given = [f.type for f in fields(cls)]
    if types == types_given:
        return AnnotationLevel.CORRECT
    return AnnotationLevel.WRONG


def has_annotation_method(f: Callable, param_types: list[object], return_type: object = None) -> AnnotationLevel:
    a = f.__annotations__
    if len(a) == 0:
        return AnnotationLevel.NONE

    ret = a['return'] if 'return' in a else None
    if ret != return_type:
        return AnnotationLevel.WRONG

    self_a = a['self'] if 'self' in a else f.__qualname__.split('.')[0]
    if self_a != f.__qualname__.split('.')[0]:
        return AnnotationLevel.WRONG

    params = [a[p] for p in a if p != 'return' and p != 'self']
    if params != param_types:
        return AnnotationLevel.WRONG
    return AnnotationLevel.CORRECT


def has_annotation_lambda(module, f_name: str, param_types: list[object], return_type: object = None) -> AnnotationLevel:
    annotation = module.__annotations__[f_name]

    if annotation.__name__ != 'Callable' and len(annotation.__args__) == 0:
        return AnnotationLevel.NONE
    if annotation.__name__ != 'Callable':
        return AnnotationLevel.WRONG

    all_types = annotation.__args__
    types, ret_type = all_types[:-1], all_types[-1]
    if ret_type != return_type:
        return AnnotationLevel.WRONG

    for param, param_type in zip(types, param_types):
        while isinstance(param, TypeAliasType):
            param = param.__value__
        while isinstance(param_type, TypeAliasType):
            param_type = param_type.__value__
        if param != param_type:
            return AnnotationLevel.WRONG
    return AnnotationLevel.CORRECT


def has_attributes(o: object, inherited_attr: list[str], new_attr: list[str]):
    attr = [f.name for f in fields(o)]
    if attr != inherited_attr + new_attr:
        return False

    for attr in inherited_attr:
        if attr in o.__annotations__:
            return False
    return True


def function_used(f: Callable, name: str) -> bool:
    source = inspect.getsource(f)
    if source.startswith(4 * " "):
        source = "\n".join(l[4:] for l in source.split("\n"))
    return name in [c.func.id for c in ast.walk(ast.parse(source)) if isinstance(c, ast.Call) and isinstance(c.func, ast.Name)]


def list_comprehension_used(f: Callable) -> bool:
    source = inspect.getsource(f)
    tree = ast.parse(source)

    for node in ast.walk(tree):
        if not isinstance(node, (ast.ListComp, ast.SetComp, ast.DictComp, ast.GeneratorExp)):
            return False
    return True


def function_not_used(f: Callable, name: str) -> bool:
    return not function_used(f, name)


def list_comprehension_not_used(f: Callable) -> bool:
    return not list_comprehension_used(f)


def has_side_effects(f: Callable, *args) -> bool:
    args_copy = deepcopy(args)
    f(*args)
    return args == args_copy


def has_no_return(f: Callable, *args) -> bool:
    args_copy = deepcopy(args)
    return f(*args_copy) is None

def is_enum(cls: object) -> bool:
    return issubclass(cls, Enum) and not is_dataclass(cls)


def uses_pattern_matching(f: Callable) -> bool:
    source = inspect.getsource(f)
    tree = ast.parse(source)
    return any(isinstance(node, ast.Match) for node in ast.walk(tree))


def is_gen(g: Iterator, res: Iterable, check_end: bool = True, max: int = 100) -> bool:
    if not isinstance(g, Iterator):
        return False
    for i, x in enumerate(res):
        try:
            el = next(g)
            if el != x:
                return False
        except StopIteration:
            return False
        if i >= max:
            break
    if not check_end:
        return True
    try:
        next(g)
    except StopIteration:
        return True
    return False


def __remove_docstrings(lines: list[str]) -> list[str]:
    new_lines = []
    i = 0
    while i < len(lines):
        line = lines[i].strip()
        if not line.startswith('"'):
            new_lines.append(line)
            i += 1
            continue
        j = 0
        while j + 1 < len(line) and line[j + 1] == '"':
            j += 1
        prefix = line[:j + 1]
        i += 1
        if line.endswith(prefix):
            continue
        while i < len(lines) and prefix not in lines[i]:
            i += 1
    return new_lines


def is_oneliner(f: Callable) -> bool:
    if f.__name__ == '<lambda>':
        return
    inspect.getsource(f)
    lines = [line for line in inspect.getsource(f).split(
        '\n') if not line.strip().startswith('#') and line.strip() != '']
    return len(__remove_docstrings(lines)) == 2
